# coding=utf-8

import sys
import argparse

from .errors import UsageError
from . import commands
from .commands import Command


def _iter_command_classes():
    for obj in vars(commands).values():
        if isinstance(obj, type) and issubclass(obj, Command) and obj is not Command:
            yield obj


def _get_commands_from_module():
    d = {}
    for cmd in _iter_command_classes():
        o = cmd()
        if o.name:
            d[o.name] = o
    return d


def _print_commands():
    print("Usage: gspider <command> [options] [args]\n")
    print("Available commands:")
    cmds = _get_commands_from_module()
    for cmdname, cmdclass in sorted(cmds.items()):
        print("  {:<10} {}".format(cmdname, cmdclass.short_desc))
    print()
    print('Use "gspider <command> -h" to see more info about a command')


def _print_unknown_command(cmdname):
    print("Unknown command: %s\n" % cmdname, file=sys.stderr)
    print('Use "gspider" to see available commands', file=sys.stderr)


def main(argv=None):
    if argv is None:
        argv = sys.argv
    cmds = _get_commands_from_module()
    cmdname = argv[1] if len(argv) > 1 else None
    if not cmdname:
        _print_commands()
        sys.exit(0)
    elif cmdname not in cmds:
        _print_unknown_command(cmdname)
        sys.exit(2)
    del argv[1]
    cmd = cmds[cmdname]
    parser = argparse.ArgumentParser()
    parser.usage = "gspider {} {}".format(cmdname, cmd.syntax)
    parser.description = cmd.long_desc
    cmd.add_arguments(parser)
    try:
        args = parser.parse_args(args=argv[1:])
        cmd.process_arguments(args)
        cmd.run(args)
    except UsageError as e:
        if e.print_help:
            parser.print_help(sys.stderr)
        print('Error: {}'.format(e), file=sys.stderr)
        sys.exit(2)
    else:
        if cmd.exitcode:
            sys.exit(cmd.exitcode)



# coding=utf-8

import logging
import os
from os.path import join, isfile
import sys
import signal

from .config import Config, DEFAULT_CONFIG
from .crawler import CrawlerRunner, Crawler
from .utils import configure_logger, daemonize, load_config, iter_settings
from .spider import RequestsSpider

log = logging.getLogger(__name__)


def run_spider(spider, **kwargs):
    kwargs['spider'] = spider
    run_crawler(config=kwargs)


def run_crawler(proj_dir=None, config=None):
    _c = Config(DEFAULT_CONFIG)
    _c.update(load_project_config(proj_dir))
    _c.update(config)
    config = _c

    configure_logger('gspider', level=config.get('log_level'), file=config.get('log_file'))
    if config.getbool('daemon'):
        daemonize()
    pid_file = config.get('pid_file')
    _write_pid_file(pid_file)
    try:
        crawler_runner = CrawlerRunner(Crawler(config))
    except Exception:
        log.error('Failed to create crawler', exc_info=True)
        _remove_pid_file(pid_file)
        raise
    default_signal_handlers = _set_signal_handlers(crawler_runner)
    try:
        crawler_runner.run()
    finally:
        _remove_pid_file(pid_file)
        _recover_signal_handlers(default_signal_handlers)


def make_requests(requests, **kwargs):
    if 'log_level' not in kwargs:
        kwargs['log_level'] = 'WARNING'
    start_requests = [r for r in requests]
    results = [None] * len(start_requests)
    run_spider(RequestsSpider, start_requests=start_requests, results=results, **kwargs)
    return results


def load_project_config(proj_dir):
    if proj_dir is not None and proj_dir not in sys.path:
        # add project path
        sys.path.append(proj_dir)
    config = Config()
    if proj_dir is not None:
        config_file = join(proj_dir, 'config.py')
        if isfile(config_file):
            c = load_config(config_file)
            for k, v in iter_settings(c):
                config[k] = v
    return config


def _write_pid_file(pid_file):
    if pid_file is not None:
        with open(pid_file, 'w') as f:
            f.write(str(os.getpid()))


def _remove_pid_file(pid_file):
    if pid_file is not None:
        try:
            os.remove(pid_file)
        except Exception as e:
            log.error('Cannot remove PID file %s: %s', pid_file, e)


def _set_signal_handlers(crawler):
    def _exit(signum, frame):
        log.info('Received exit signal: %s', signum)
        crawler.stop()

    default_signal_handlers = [(signal.SIGINT, signal.getsignal(signal.SIGINT)),
                               (signal.SIGTERM, signal.getsignal(signal.SIGTERM))]
    signal.signal(signal.SIGINT, _exit)
    signal.signal(signal.SIGTERM, _exit)
    return default_signal_handlers


def _recover_signal_handlers(handlers):
    for h in handlers:
        signal.signal(h[0], h[1])



# coding=utf-8

import copy
from collections import MutableMapping


class Config(MutableMapping):
    def __init__(self, __values=None, **kwargs):
        self.attrs = {}
        self.update(__values, **kwargs)

    def __getitem__(self, name):
        if name not in self:
            return None
        return self.attrs[name]

    def __contains__(self, name):
        return name in self.attrs

    def get(self, name, default=None):
        return self[name] if self[name] is not None else default

    def getbool(self, name, default=None):
        v = self.get(name, default)
        return getbool(v)

    def getint(self, name, default=None):
        v = self.get(name, default)
        return getint(v)

    def getfloat(self, name, default=None):
        v = self.get(name, default)
        return getfloat(v)

    def getlist(self, name, default=None):
        v = self.get(name, default)
        return getlist(v)

    def __setitem__(self, name, value):
        self.attrs[name] = value

    def set(self, name, value):
        self[name] = value

    def setdefault(self, k, default=None):
        if k not in self:
            self[k] = default

    def update(self, __values=None, **kwargs):
        if __values is not None:
            if isinstance(__values, Config):
                for name in __values:
                    self[name] = __values[name]
            else:
                for name, value in __values.items():
                    self[name] = value
        for k, v in kwargs.items():
            self[k] = v

    def delete(self, name):
        del self.attrs[name]

    def __delitem__(self, name):
        del self.attrs[name]

    def copy(self):
        return copy.deepcopy(self)

    def __iter__(self):
        return iter(self.attrs)

    def __len__(self):
        return len(self.attrs)


def getbool(v):
    try:
        return bool(int(v))
    except (ValueError, TypeError):
        if v in ("True", "true"):
            return True
        if v in ("False", "false"):
            return False
    return None


def getint(v):
    try:
        return int(v)
    except (ValueError, TypeError):
        pass
    return None


def getfloat(v):
    try:
        return float(v)
    except (ValueError, TypeError):
        pass
    return None


def getlist(v):
    if v is None:
        return None
    if isinstance(v, str):
        v = v.split(",")
    elif not hasattr(v, "__iter__"):
        v = [v]
    return list(v)


DEFAULT_CONFIG = {
    'daemon': False,
    'log_level': 'info',
    'log_file': None,
    'fetcher': 'gspider.fetcher.Fetcher',
    'queue': 'gspider.queue.PriorityQueue',
    'dupe_filter': 'gspider.dupefilter.HashDupeFilter',
    'default_extensions': [
        'gspider.extensions.RetryMiddleware',
        'gspider.extensions.DepthMiddleware',
    ],
    'max_workers': 100,
    'default_headers': None,
    'verify_ssl': None,
    'proxies': None,
    'max_retry_times': None,
    'max_depth': None,
}



# coding=utf-8

from os.path import abspath, isfile, isdir, split, splitext
import logging
from importlib import import_module
import sys
import inspect

from .errors import UsageError
from .utils import load_config, iter_settings
from . import __version__
from .run import run_crawler
from .spider import Spider

log = logging.getLogger(__name__)


class Command:
    def __init__(self):
        self.exitcode = 0

    @property
    def name(self):
        return ""

    @property
    def syntax(self):
        return ""

    @property
    def short_desc(self):
        return ""

    @property
    def long_desc(self):
        return self.short_desc

    def add_arguments(self, parser):
        pass

    def process_arguments(self, args):
        pass

    def run(self, args):
        raise NotImplementedError


def _import_spider(file):
    file = abspath(file)
    d, f = split(file)
    m, ext = splitext(f)
    if ext != '.py':
        raise UsageError('{} is not a python module'.format(file))
    if d not in sys.path:
        sys.path.append(d)
    module = import_module(m)
    for v in vars(module).values():
        if inspect.isclass(v) and issubclass(v, Spider) and v.__module__ == module.__name__:
            return v
    raise UsageError('Cannot find spider in {}'.format(file))


class Option:
    def __init__(self, name=None, cli=None, metavar=None, default=None, action=None, type=None, nargs=None,
                 short_desc=None):
        self.name = name
        self.cli = cli
        self.metavar = metavar
        self.default = default
        self.action = action
        self.type = type
        self.nargs = nargs
        self.short_desc = short_desc

    def add_argument(self, parser):
        if self.cli is None:
            return
        args = tuple(self.cli)
        kwargs = {'dest': self.name, 'help': self.short_desc}
        if self.metavar is not None:
            kwargs['metavar'] = self.metavar
        if self.action is not None:
            kwargs['action'] = self.action
        if self.type is not None:
            kwargs['type'] = self.type
        if self.nargs is not None:
            kwargs['nargs'] = self.nargs
        parser.add_argument(*args, **kwargs)


class CrawlCommand(Command):
    def __init__(self):
        super().__init__()
        self.config = {}
        self.options = [
            Option(name='daemon', cli=['-d', '--daemon'], action='store_true', short_desc='run in daemon mode'),
            Option(name='log_level', cli=['-l', '--log-level'], metavar='LEVEL', short_desc='log level'),
            Option(name='log_file', cli=['--log-file'], metavar='FILE', short_desc='log file'),
            Option(name='pid_file', cli=['--pid-file'], metavar='FILE', short_desc='PID file')]

    @property
    def syntax(self):
        return "[options] <PATH>"

    @property
    def name(self):
        return "crawl"

    @property
    def short_desc(self):
        return "Start to crawl web pages"

    def add_arguments(self, parser):
        parser.add_argument("path", metavar="PATH", nargs=1, help="project directory or spider file")
        parser.add_argument('-c', '--config', dest='config', metavar='FILE',
                            help='configuration file')
        for s in self.options:
            s.add_argument(parser)
        parser.add_argument("-s", "--set", dest="set", action="append", default=[], metavar="NAME=VALUE",
                            help="set/override setting (can be repeated)")

    def process_arguments(self, args):
        args.path = args.path[0]
        if args.config is not None:
            c = load_config(args.config)
            for k, v in iter_settings(c):
                self.config[k] = v
        try:
            self.config.update(dict(x.split("=", 1) for x in args.set))
        except ValueError:
            raise UsageError("Invalid -s value, use -s NAME=VALUE")
        for s in self.options:
            v = getattr(args, s.name)
            if v is not None:
                self.config[s.name] = v

    def run(self, args):
        if isfile(args.path):
            spider = _import_spider(args.path)
            self.config['spider'] = spider
            run_crawler(proj_dir=None, config=self.config)
        elif isdir(args.path):
            run_crawler(proj_dir=args.path, config=self.config)
        else:
            raise UsageError('Cannot find {}'.format(args.path))


class VersionCommand(Command):
    @property
    def name(self):
        return "version"

    @property
    def short_desc(self):
        return "Print the version"

    def run(self, args):
        print("gspider version {}".format(__version__))



# coding=utf-8

from gspider import Spider, HttpRequest, run_spider, Selector


class BaiduNewsSpider(Spider):
    def start_requests(self):
        yield HttpRequest("http://news.baidu.com/")

    def parse(self, response):
        selector = Selector(response.text)
        hot = selector.css("div.hotnews a").text
        self.log("Hot News:")
        for i in range(len(hot)):
            self.log("%s: %s", i + 1, hot[i])


if __name__ == '__main__':
    run_spider(BaiduNewsSpider)



# coding=utf-8

import logging

from .utils import load_object, isiterable
from . import events
from .errors import NotEnabled
from .http import HttpRequest, HttpResponse

log = logging.getLogger(__name__)


class ExtensionManager:
    def __init__(self, *extensions):
        self.extensions = []
        for ext in extensions:
            self._add_extension(ext)

    @classmethod
    def from_crawler(cls, crawler):
        ext_list = cls._extension_list_from_config(crawler.config)
        exts = []
        for cls_path in ext_list:
            ext_cls = load_object(cls_path)
            try:
                if hasattr(ext_cls, "from_crawler"):
                    ext = ext_cls.from_crawler(crawler)
                else:
                    ext = ext_cls()
            except NotEnabled:
                log.debug('%s is not enabled', cls_path)
            else:
                exts.append(ext)
        obj = cls(*exts)
        crawler.event_bus.subscribe(obj.open, events.crawler_start)
        crawler.event_bus.subscribe(obj.close, events.crawler_shutdown)
        return obj

    def _add_extension(self, ext):
        assert isinstance(ext, Extension)
        self.extensions.append(ext)

    @classmethod
    def _extension_list_from_config(cls, config):
        return cls._make_component_list('extensions', config)

    @staticmethod
    def _list_from_config(name, config):
        c = config.get(name)
        assert c is None or isinstance(c, list), \
            "'{}' must be None or a list, got {}".format(name, type(c).__name__)
        if c is None:
            return []
        return c

    @classmethod
    def _make_component_list(cls, name, config):
        c_base = cls._list_from_config('default_' + name, config)
        c = cls._list_from_config(name, config)
        return c + c_base

    def open(self):
        for ext in self.extensions:
            ext.open()

    def close(self):
        for ext in self.extensions:
            ext.close()

    def handle_request(self, request):
        for ext in self.extensions:
            res = ext.handle_request(request)
            assert res is None or isinstance(res, (HttpRequest, HttpResponse)), \
                "Request handler must return None, HttpRequest or HttpResponse, got {}".format(type(res).__name__)
            if res:
                return res

    def handle_response(self, request, response):
        for ext in self.extensions:
            res = ext.handle_response(request, response)
            assert res is None or isinstance(res, HttpRequest), \
                "Response handler must return None or HttpRequest, got {}".format(type(res).__name__)
            if res:
                return res

    def handle_error(self, request, error):
        for ext in self.extensions:
            res = ext.handle_error(request, error)
            assert res is None or isinstance(res, (HttpRequest, HttpResponse)), \
                "Exception handler must return None, HttpRequest or HttpResponse, got {}".format(type(res).__name__)
            if res:
                return res
        return error

    def handle_spider_input(self, response):
        for ext in self.extensions:
            res = ext.handle_spider_input(response)
            assert res is None, \
                "Spider input handler must return None, got {}".format(type(res).__name__)

    def handle_spider_output(self, response, result):
        for ext in self.extensions:
            result = ext.handle_spider_output(response, result)
            assert isiterable(result), \
                "Spider output handler must return an iterable object, got {}".format(type(result).__name__)
        return result

    def handle_spider_error(self, response, error):
        for ext in self.extensions:
            res = ext.handle_spider_error(response, error)
            assert res is None or isiterable(res), \
                "Spider exception handler must return None or an iterable object, got {}".format(type(res).__name__)
            if res is not None:
                return res
        return error

    def handle_start_requests(self, result):
        for ext in self.extensions:
            result = ext.handle_start_requests(result)
            assert isiterable(result), \
                "Start requests handler must return an iterable object, got {}".format(type(result).__name__)
        return result


class Extension:
    def open(self):
        pass

    def close(self):
        pass

    def handle_request(self, request):
        pass

    def handle_response(self, request, response):
        pass

    def handle_error(self, request, error):
        pass

    def handle_spider_input(self, response):
        pass

    def handle_spider_output(self, response, result):
        return result

    def handle_spider_error(self, response, error):
        pass

    def handle_start_requests(self, result):
        return result



# coding=utf-8

from . import _patch
from .http import HttpRequest, HttpResponse
from .fetcher import Fetcher
from .spider import Spider
from .selector import Selector
from .run import run_spider, make_requests
from .errors import StopCrawler

__all__ = ['HttpRequest', 'HttpResponse',
           'Fetcher',
           'Spider',
           'Selector',
           'run_spider', 'make_requests',
           'StopCrawler']

__version__ = '0.1.0'

del _patch



# coding=utf-8


class NotEnabled(Exception):
    """
    Not enabled.
    """


class UsageError(Exception):
    """
    Command usage error.
    """

    def __init__(self, *args, print_help=False, **kwargs):
        self.print_help = print_help
        super().__init__(*args, **kwargs)


class ClientError(Exception):
    """
    Client error.
    """


class IgnoreRequest(Exception):
    """
    Ignore this request.
    """


class HttpError(Exception):
    """
    HTTP status is not 2xx.
    """

    def __init__(self, *args, response=None, **kwargs):
        self.response = response
        super().__init__(*args, **kwargs)


class StopCrawler(Exception):
    """
    Stop crawler.
    """



# coding=utf-8

import weakref
import logging

log = logging.getLogger(__name__)


class EventBus:
    def __init__(self):
        self._refs = {}

    def subscribe(self, receiver, event):
        if event not in self._refs:
            self._refs[event] = {}
        if not hasattr(receiver, '__func__') or not hasattr(receiver, '__self__'):
            raise ValueError("Failed to subscribe, {} has no attribute '__func__' or '__self__'".format(receiver))
        i = self._calc_id(receiver)
        if i in self._refs[event]:
            f = self._refs[event][i]()
            if f is not None:
                return
        self._refs[event][i] = weakref.WeakMethod(receiver)

    def unsubscribe(self, receiver, event):
        if event in self._refs:
            i = self._calc_id(receiver)
            if i in self._refs[event]:
                del self._refs[event][i]

    def send(self, event, **kwargs):
        if event not in self._refs:
            return
        del_list = []
        for i in self._refs[event]:
            f = self._refs[event][i]()
            if f is None:
                del_list.append(i)
            else:
                try:
                    f(**kwargs)
                except Exception:
                    log.error("Failed to send the event", exc_info=True)
        for i in del_list:
            del self._refs[event][i]
        del del_list

    def _calc_id(self, receiver):
        return hash((id(receiver.__func__), id(receiver.__self__)))



# coding=utf-8

import time
import logging
from collections import deque
from heapq import heappush, heappop

from gevent.lock import Semaphore

from .utils import cmp

log = logging.getLogger(__name__)


class FifoQueue:
    def __init__(self):
        self._queue = deque()
        self._semaphore = Semaphore(0)

    def __len__(self):
        return len(self._queue)

    def push(self, request):
        self._queue.append(request)
        self._semaphore.release()

    def pop(self):
        self._semaphore.acquire()
        return self._queue.popleft()


class LifoQueue(FifoQueue):
    def pop(self):
        self._semaphore.acquire()
        return self._queue.pop()


class PriorityQueue:
    def __init__(self):
        self._queue = []
        self._semaphore = Semaphore(0)

    def __len__(self):
        return len(self._queue)

    def push(self, request):
        heappush(self._queue, _PriorityQueueItem(request))
        self._semaphore.release()

    def pop(self):
        self._semaphore.acquire()
        item = heappop(self._queue)
        return item.request


class _PriorityQueueItem:
    def __init__(self, request):
        self.request = request
        self.priority = self.request.priority or 0
        self.now = time.time()

    def __cmp__(self, other):
        return cmp((-self.priority, self.now), (-other.priority, other.now))

    def __lt__(self, other):
        return self.__cmp__(other) < 0



# coding=utf-8

import logging

from .utils import request_fingerprint

log = logging.getLogger(__name__)


class HashDupeFilter:
    def __init__(self):
        self._hash = set()

    def is_duplicated(self, request):
        if request.dont_filter:
            return False
        h = request_fingerprint(request)
        if h in self._hash:
            log.debug("%s is duplicated", request)
            return True
        self._hash.add(h)
        return False

    def clear(self):
        self._hash.clear()



# coding=utf-8

import logging

from gspider.errors import ClientError, HttpError
from gspider.extension import Extension

log = logging.getLogger(__name__)

__all__ = ['RetryMiddleware']


class RetryMiddleware(Extension):
    RETRY_ERRORS = (ClientError,)
    RETRY_HTTP_STATUS = (500, 502, 503, 504, 408, 429)

    def __init__(self, max_retry_times=3, retry_http_status=None):
        self._max_retry_times = max_retry_times
        self._retry_http_status = retry_http_status or self.RETRY_HTTP_STATUS

    def __repr__(self):
        cls_name = self.__class__.__name__
        return '{}(max_retry_times={}, retry_http_status={})' \
            .format(cls_name, repr(self._max_retry_times), repr(self._retry_http_status))

    @classmethod
    def from_crawler(cls, crawler):
        config = crawler.config
        max_retry_times = config.getint('max_retry_times')
        retry_http_status = config.getlist('retry_http_status')
        kwargs = {}
        if max_retry_times is not None:
            assert max_retry_times >= 0, 'max retry times should >= 0'
            kwargs['max_retry_times'] = max_retry_times
        if retry_http_status is not None:
            kwargs['retry_http_status'] = retry_http_status
        return cls(**kwargs)

    def handle_response(self, request, response):
        for p in self._retry_http_status:
            if self.match_status(p, response.status):
                return self.retry(request, "HTTP status={}".format(response.status))

    @staticmethod
    def match_status(pattern, status):
        if isinstance(pattern, int):
            return pattern == status
        verse = False
        if pattern.startswith("!") or pattern.startswith("~"):
            verse = True
            pattern = pattern[1:]
        s = str(status)
        n = len(s)
        match = True
        if len(pattern) != n:
            match = False
        else:
            for i in range(n):
                if pattern[i] != "x" and pattern[i] != "X" and pattern[i] != s[i]:
                    match = False
                    break
        if verse:
            match = not match
        return match

    def handle_error(self, request, error):
        if isinstance(error, self.RETRY_ERRORS):
            return self.retry(request, error)
        if isinstance(error, HttpError):
            for p in self._retry_http_status:
                if self.match_status(p, error.response.status):
                    return self.retry(request, "HTTP status={}".format(error.response.status))

    def retry(self, request, reason):
        retry_times = request.meta.get('retry_times', 0) + 1
        if retry_times <= self._max_retry_times:
            log.debug('Retry %s (failed %s times): %s', request, retry_times, reason)
            retry_req = request.copy()
            retry_req.meta['retry_times'] = retry_times
            retry_req.dont_filter = True
            return retry_req
        else:
            log.debug('Give up retrying %s (failed %s times): %s', request, retry_times, reason)



# coding=utf-8

from gspider.spider import Spider
from gspider.http import HttpRequest
from gspider.run import run_spider
from gspider.extension import Extension


class FooError(Exception):
    pass


class HandlerDownloaderMiddleware(Extension):
    def handle_request(self, request):
        if request.url.endswith('error'):
            raise FooError


class HandlerSpiderMiddleware(Extension):
    def handle_spider_input(self, response):
        if response.request.url.endswith('not-found'):
            raise FooError

    def handle_spider_error(self, response, error):
        if isinstance(error, FooError):
            return ()


class HandlerSpider(Spider):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data = self.config.get('data')
        self.server_address = self.config.get('server_address')

    def start_requests(self):
        yield HttpRequest("http://unknown/", errback=self.error_back)
        yield HttpRequest("http://{}/error".format(self.server_address), errback=self.handle_request_error)
        yield HttpRequest("http://{}/".format(self.server_address), dont_filter=True)
        yield HttpRequest("http://{}/".format(self.server_address), dont_filter=True, callback=self.generator_parse)
        yield HttpRequest("http://{}/".format(self.server_address), dont_filter=True, callback=self.func_prase)
        yield HttpRequest("http://{}/".format(self.server_address), dont_filter=True, callback=self.return_list_parse)
        yield HttpRequest("http://{}/".format(self.server_address), dont_filter=True, callback=self.return_none_parse)

    def parse(self, response):
        self.data.add('parse')

    def error_back(self, request, err):
        self.data.add('error_back')
        raise RuntimeError('not an error actually')

    def handle_request_error(self, request, error):
        assert isinstance(error, FooError)
        self.data.add('handle_request_error')

    def generator_parse(self, response):
        self.data.add('generator_parse')
        if response.status / 100 != 2:
            raise RuntimeError('not an error actually')
        # it will never come here
        yield None

    def func_prase(self, response):
        self.data.add('func_parse')
        raise RuntimeError('not an error actually')

    def return_list_parse(self, response):
        self.data.add('return_list_parse')
        return []

    def return_none_parse(self, response):
        self.data.add('return_none_parse')


def test_spider_handlers():
    data = set()
    run_spider(HandlerSpider, log_level='DEBUG', extensions=[HandlerDownloaderMiddleware, HandlerSpiderMiddleware],
               data=data, server_address='python.org')
    assert 'parse' in data
    assert 'error_back' in data
    assert 'handle_request_error' in data
    assert 'generator_parse' in data
    assert 'func_parse' in data
    assert 'return_list_parse' in data
    assert 'return_none_parse' in data



# coding=utf-8

import os
import hashlib
import logging
from importlib import import_module
from os.path import isfile

from urllib.parse import urlsplit, parse_qsl, urlencode


def load_object(path):
    if isinstance(path, str):
        dot = path.rindex(".")
        module, name = path[:dot], path[dot + 1:]
        mod = import_module(module)
        return getattr(mod, name)
    return path


default_log_level = 'info'
default_log_format = '%(asctime)s %(name)s [%(levelname)s] %(message)s'
default_log_date_format = '[%Y-%m-%d %H:%M:%S %z]'


def configure_logger(name, level=None, format=None, date_format=None, file=None):
    if level is None:
        level = default_log_level
    if format is None:
        format = default_log_format
    if date_format is None:
        date_format = default_log_date_format

    level = level.upper()
    logger = logging.getLogger(name)
    logger.setLevel(level)
    if file:
        handler = logging.FileHandler(file)
    else:
        handler = logging.StreamHandler()
    formatter = logging.Formatter(format, date_format)
    handler.setFormatter(formatter)
    logger.handlers.clear()
    logger.addHandler(handler)
    return logger


def to_bytes(data, encoding=None):
    if isinstance(data, bytes):
        return data
    if isinstance(data, str):
        return data.encode(encoding or "utf-8")
    raise TypeError("Need bytes or str, got {}".format(type(data).__name__))


def iterable_to_list(gen):
    res = []
    if gen is not None:
        for r in gen:
            res.append(r)
    return res


def daemonize():
    if os.fork():
        os._exit(0)
    os.setsid()
    if os.fork():
        os._exit(0)
    os.umask(0o22)
    os.closerange(0, 3)
    fd_null = os.open(os.devnull, os.O_RDWR)
    if fd_null != 0:
        os.dup2(fd_null, 0)
    os.dup2(fd_null, 1)
    os.dup2(fd_null, 2)


def load_config(fname):
    if fname is None or not isfile(fname):
        raise ValueError('{} is not a file'.format(fname))
    code = compile(open(fname, 'rb').read(), fname, 'exec')
    cfg = {
        "__builtins__": __builtins__,
        "__name__": "__config__",
        "__file__": fname,
        "__doc__": None,
        "__package__": None
    }
    exec(code, cfg, cfg)
    return cfg


def iter_settings(config):
    for key, value in config.items():
        if not key.startswith('_'):
            yield key, value


def isiterable(obj):
    return hasattr(obj, "__iter__")


def cmp(a, b):
    return (a > b) - (a < b)


def request_fingerprint(request):
    sha1 = hashlib.sha1()
    sha1.update(to_bytes(request.method))
    res = urlsplit(request.url)
    queries = parse_qsl(res.query)
    queries.sort()
    final_query = urlencode(queries)
    sha1.update(to_bytes('{}://{}{}:{}?{}'.format(res.scheme,
                                                  '' if res.hostname is None else res.hostname,
                                                  res.path,
                                                  80 if res.port is None else res.port,
                                                  final_query)))
    sha1.update(request.body or b'')
    return sha1.hexdigest()



# coding=utf-8

import json

import pytest

from gspider.http import HttpRequest
from gspider.fetcher import Fetcher
from gspider.errors import HttpError


def test_basic_auth():
    fetcher = Fetcher()

    def no_auth():
        req = HttpRequest("http://httpbin.org/basic-auth/user/passwd")
        with pytest.raises(HttpError) as e:
            fetcher.fetch(req)
            assert e.value.response.status == 401

    def tuple_auth():
        req = HttpRequest("http://httpbin.org/basic-auth/user/passwd")
        req.auth = ('user', 'passwd')
        resp = fetcher.fetch(req)
        assert resp.status == 200

    no_auth()
    tuple_auth()


def test_params():
    fetcher = Fetcher()

    def query_params():
        url = "http://httpbin.org/anything?key=value&none="
        resp = fetcher.fetch(HttpRequest(url))
        assert json.loads(resp.text)['args'] == {'key': 'value', 'none': ''}

    def dict_params():
        resp = fetcher.fetch(
            HttpRequest("http://httpbin.org/get", params={'key': 'value', 'none': ''}))
        assert json.loads(resp.text)['args'] == {'key': 'value', 'none': ''}

    def list_params():
        resp = fetcher.fetch(HttpRequest("http://httpbin.org/get",
                                         params=[('list', '1'), ('list', '2')]))
        assert json.loads(resp.text)['args'] == {'list': ['1', '2']}

    query_params()
    dict_params()
    list_params()


def test_headers():
    fetcher = Fetcher()
    headers = {'User-Agent': 'gspider'}
    resp = fetcher.fetch(HttpRequest("http://httpbin.org/get",
                                     headers=headers))
    assert resp.status == 200
    data = json.loads(resp.text)['headers']
    assert 'User-Agent' in data and data['User-Agent'] == 'gspider'


def test_body():
    fetcher = Fetcher()

    def post_bytes():
        bytes_data = 'bytes data: 字节数据'
        resp = fetcher.fetch(HttpRequest('http://httpbin.org/post',
                                         'POST', body=bytes_data.encode(),
                                         headers={'Content-Type': 'text/plain'}))
        assert resp.status == 200
        body = json.loads(resp.text)['data']
        assert body == bytes_data

    post_bytes()


def test_allow_redirects():
    fetcher = Fetcher()

    resp = fetcher.fetch(HttpRequest('http://httpbin.org/redirect-to',
                                     params={'url': 'http://python.org'}))
    assert resp.status // 100 == 2 and 'python.org' in resp.url

    resp = fetcher.fetch(HttpRequest('http://httpbin.org/redirect-to',
                                     params={'url': 'http://python.org'},
                                     allow_redirects=False))
    assert resp.status // 100 == 3



# coding=utf-8

import logging
import inspect

import gevent

from .http import HttpRequest, HttpResponse
from .errors import IgnoreRequest, StopCrawler, ClientError, HttpError
from .spider import Spider
from .eventbus import EventBus
from . import events
from .extension import ExtensionManager
from .utils import load_object, iterable_to_list, isiterable

log = logging.getLogger(__name__)


class Crawler:
    def __init__(self, config):
        self.config = config
        self.event_bus = EventBus()
        self.queue = self._instance_from_crawler(self.config.get('queue'))
        self.dupe_filter = self._instance_from_crawler(self.config.get('dupe_filter'))
        self.fetcher = self._instance_from_crawler(self.config.get('fetcher'))
        self.spider = self._instance_from_crawler(self.config.get('spider'))
        assert isinstance(self.spider, Spider), 'spider must inherit from the Spider class'
        log.info('Spider class: %s', self.spider.__class__.__name__)
        self.extension = ExtensionManager.from_crawler(self)
        log.info('Extensions: %s', self._log_objects(self.extension.extensions))

    def start_requests(self):
        try:
            res = self.spider.start_requests()
            assert res is None or isiterable(res), \
                "Start requests must be None or an iterable object, got {}".format(type(res).__name__)
            result = iterable_to_list(res)
            if result:
                res = self.extension.handle_start_requests(result)
                result = iterable_to_list(res)
        except StopCrawler:
            raise
        except Exception:
            log.error("Failed to get start requests", exc_info=True)
        else:
            return result

    def schedule(self, request):
        try:
            res = self.dupe_filter.is_duplicated(request)
            if not res:
                self.event_bus.send(events.request_scheduled, request=request)
                self.queue.push(request)
        except StopCrawler:
            raise
        except Exception:
            log.error('Failed to schedule %s', request, exc_info=True)

    def next_request(self):
        req = self.queue.pop()
        return req

    def fetch(self, req):
        try:
            resp = self._fetch(req)
        except StopCrawler:
            raise
        except Exception as e:
            if isinstance(e, IgnoreRequest):
                self.event_bus.send(events.request_ignored, request=req, error=e)
            elif isinstance(e, (ClientError, HttpError)):
                log.info('Failed to make %s: %s', req, e)
            else:
                log.warning("Failed to request %s", req, exc_info=True)
            self.spider.handle_error(req, e)
        else:
            self._handle_response(resp)

    def _fetch(self, req):
        try:
            res = self.extension.handle_request(req)
            if isinstance(res, HttpRequest):
                return res
            if res is None:
                res = self.fetcher.fetch(req)
        except StopCrawler:
            raise
        except Exception as e:
            res = self.extension.handle_error(req, e)
            if isinstance(res, Exception):
                raise res
        if isinstance(res, HttpResponse):
            _res = self.extension.handle_response(req, res)
            if _res:
                res = _res
            # bind request
            res.request = req
        return res

    def _handle_response(self, resp):
        if isinstance(resp, HttpRequest):
            self.schedule(resp)
        elif isinstance(resp, HttpResponse):
            self.event_bus.send(events.response_received, response=resp)
            try:
                result = self._parse(resp)
            except StopCrawler:
                raise
            except Exception as e:
                if isinstance(e, IgnoreRequest):
                    self.event_bus.send(events.request_ignored, request=resp.request, error=e)
                else:
                    log.warning("Failed to parse %s", resp, exc_info=True)
            else:
                for r in result:
                    self._handle_parsing_result(r)

    def _parse(self, response):
        request = response.request
        try:
            try:
                self.extension.handle_spider_input(response)
            except Exception as e:
                self.spider.handle_error(request, e)
                raise e
            res = self.spider.handle_response(response)
            assert res is None or isiterable(res), \
                "Parsing result must be None or an iterable object, got {}".format(type(res).__name__)
            result = iterable_to_list(res)
        except Exception as e:
            res = self.extension.handle_spider_error(response, e)
            if isinstance(res, Exception):
                raise res
            result = iterable_to_list(res)
        if result:
            res = self.extension.handle_spider_output(response, result)
            result = iterable_to_list(res)
        return result

    def _handle_parsing_result(self, result):
        if isinstance(result, HttpRequest):
            self.schedule(result)

    def _instance_from_crawler(self, cls_path):
        obj_cls = load_object(cls_path)
        if inspect.isclass(obj_cls):
            if hasattr(obj_cls, "from_crawler"):
                obj = obj_cls.from_crawler(self)
            else:
                obj = obj_cls()
        else:
            obj = obj_cls
        return obj

    @staticmethod
    def _log_objects(objects):
        if objects:
            return ''.join(['\n\t({}/{}) {}'.format(i + 1, len(objects), o) for i, o in enumerate(objects)])
        return ''


class CrawlerRunner:
    def __init__(self, crawler):
        self.crawler = crawler

        self._workers = None
        self._req_in_worker = None
        self._start_requests_generator = None
        self._is_running = False

    def run(self):
        if self._is_running:
            return
        self._is_running = True

        self.crawler.event_bus.send(events.crawler_start)
        max_workers = self.crawler.config.getint('max_workers')
        assert max_workers > 0, 'max workers should > 0'
        log.info("The maximum number of workers: %s", max_workers)

        self._start_requests_generator = gevent.spawn(self._schedule_start_requests)
        self._workers = []
        for i in range(max_workers):
            self._workers.append(gevent.spawn(self._fetch, i))
        self._req_in_worker = [None] * max_workers

        log.info('Crawler is running')
        gevent.joinall([self._start_requests_generator] + self._workers)
        self.crawler.event_bus.send(events.crawler_shutdown)
        log.info('Crawler is stopped')

        self._start_requests_generator = None
        self._workers = None
        self._req_in_worker = None

    def stop(self):
        if not self._is_running:
            return
        self._is_running = False
        self._shutdown()

    def _shutdown(self):
        log.info("Shutdown now")
        self._start_requests_generator.kill(exception=StopCrawler, block=False)
        for w in self._workers:
            w.kill(exception=StopCrawler, block=False)

    def _all_done(self):
        if self._start_requests_generator.ready() and len(self.crawler.queue) <= 0:
            no_active = True
            for i in range(len(self._workers)):
                if self._req_in_worker[i]:
                    no_active = False
                    break
            return no_active
        return False

    def _schedule_start_requests(self):
        try:
            reqs = self.crawler.start_requests()
            for r in reqs:
                self.crawler.schedule(r)
        except StopCrawler:
            pass

    def _fetch(self, coro_id):
        try:
            while True:
                req = self.crawler.next_request()
                log.debug("%s -> worker[%s]", req, coro_id)
                self._req_in_worker[coro_id] = req
                try:
                    self.crawler.fetch(req)
                except StopCrawler:
                    self.stop()
                    raise
                self._req_in_worker[coro_id] = None
                # check if it's all done
                if self._all_done():
                    self.stop()
        except StopCrawler:
            pass



# coding=utf-8

import inspect

from requests.models import Response


class HttpRequest:
    def __init__(self, url, method="GET", params=None, body=None, json=None, headers=None, proxies=None,
                 timeout=20, verify_ssl=None, allow_redirects=None, auth=None,
                 priority=None, dont_filter=False, callback=None, errback=None, meta=None):
        """
        Construct an HTTP request.
        """
        self.url = url
        self.params = params
        self.method = method
        self.body = body
        self.json = json
        self.headers = headers
        self.proxies = proxies
        self.timeout = timeout
        self.verify_ssl = verify_ssl
        self.allow_redirects = allow_redirects
        self.auth = auth
        self.priority = priority
        self.dont_filter = dont_filter
        self.callback = callback
        self.errback = errback
        self._meta = dict(meta) if meta else {}

    def __str__(self):
        return '<{}, {}>'.format(self.method, self.url)

    __repr__ = __str__

    @property
    def meta(self):
        return self._meta

    def copy(self):
        return self.replace()

    def replace(self, **kwargs):
        for i in ["url", "method", "params", "body", "json", "headers", "proxies",
                  "timeout", "verify_ssl", "allow_redirects", "auth",
                  "priority", "dont_filter", "callback", "errback", "meta"]:
            kwargs.setdefault(i, getattr(self, i))
        return type(self)(**kwargs)

    def to_dict(self):
        callback = self.callback
        if inspect.ismethod(callback):
            callback = callback.__name__
        errback = self.errback
        if inspect.ismethod(errback):
            errback = errback.__name__
        d = {
            'url': self.url,
            'method': self.method,
            'body': self.body,
            'headers': self.headers,
            'proxy': self.proxy,
            'timeout': self.timeout,
            'verify_ssl': self.verify_ssl,
            'allow_redirects': self.allow_redirects,
            'auth': self.auth,
            'proxy_auth': self.proxy_auth,
            'priority': self.priority,
            'dont_filter': self.dont_filter,
            'callback': callback,
            'errback': errback,
            'meta': self.meta
        }
        return d

    @classmethod
    def from_dict(cls, d):
        return cls(**d)


class HttpResponse:
    def __init__(self, request=None, response: Response = None):
        """
        Construct an HTTP response.
        """
        self.request = request
        self.response = response

    def __str__(self):
        return '<{}, {}>'.format(self.status, self.url)

    __repr__ = __str__

    @property
    def url(self):
        if self.response:
            return self.response.url

    @property
    def status(self):
        if self.response:
            return self.response.status_code

    @property
    def body(self):
        if self.response:
            return self.response.content

    @property
    def text(self):
        if self.response:
            return self.response.text

    @property
    def meta(self):
        if self.request:
            return self.request.meta



# coding=utf-8

import logging
from abc import ABCMeta, abstractmethod

from . import events
from .http import HttpRequest

log = logging.getLogger(__name__)


class Spider(metaclass=ABCMeta):

    @classmethod
    def from_crawler(cls, crawler):
        cls.crawler = crawler
        cls.config = crawler.config
        spider = cls()
        crawler.event_bus.subscribe(spider.open, events.crawler_start)
        crawler.event_bus.subscribe(spider.close, events.crawler_shutdown)
        return spider

    @property
    def logger(self):
        return log

    def log(self, message, *args, level=logging.INFO, **kwargs):
        if isinstance(level, str):
            level = logging._nameToLevel(level.upper())
        self.logger.log(level, message, *args, **kwargs)

    @abstractmethod
    def parse(self, response):
        pass

    @abstractmethod
    def start_requests(self):
        pass

    def open(self):
        pass

    def close(self):
        pass

    def handle_response(self, response):
        callback = response.request.callback
        if callback:
            res = self._get_mothod(callback)(response)
        else:
            res = self.parse(response)
        return res

    def handle_error(self, request, error):
        try:
            if request and request.errback:
                self._get_mothod(request.errback)(request, error)
        except Exception:
            log.error("Exception encountered in the error callback of spider", exc_info=True)

    def _get_mothod(self, method):
        if isinstance(method, str):
            method = getattr(self, method)
        return method


class RequestsSpider(Spider):
    def start_requests(self):
        requests = self.config.get('start_requests')
        i = 0
        for r in requests:
            if isinstance(r, str):
                r = HttpRequest(r)
            if isinstance(r, HttpRequest):
                r.meta['request_index'] = i
                r.dont_filter = True
                r.callback = self.parse
                r.errback = self.handle_error
                yield r
            else:
                self.logger.warning('Requests must be str or HttpRequest, got %s', type(r).__name__)
            i += 1

    def parse(self, response):
        results = self.config.get('results')
        results[response.meta['request_index']] = response

    def handle_error(self, request, err):
        results = self.config.get('results')
        results[request.meta['request_index']] = err



# coding=utf-8

import logging

import requests
from requests import HTTPError, Response
import gevent

from .errors import ClientError, HttpError
from .http import HttpRequest, HttpResponse

log = logging.getLogger(__name__)


class Fetcher:
    default_session_id = '0'

    def __init__(self, default_headers=None, verify_ssl=None, proxies=None):
        self.default_headers = default_headers
        self.verify_ssl = verify_ssl
        self.proxies = proxies
        self.sessions = {}

    def new_session(self):
        session = requests.Session()
        if self.default_headers is not None:
            session.headers.update(self.default_headers)
        if self.verify_ssl is not None:
            session.verify = self.verify_ssl
        if self.proxies is not None:
            session.proxies = self.proxies
        return session

    def close_session(self, session_id):
        if session_id in self.sessions:
            session = self.sessions.pop(session_id)
            session.close()

    @classmethod
    def from_crawler(cls, crawler):
        config = crawler.config
        kwargs = {}
        if config['verify_ssl'] is not None:
            kwargs['verify_ssl'] = config['verify_ssl']
        if config['proxies'] is not None:
            kwargs['proxies'] = config['proxies']
        return cls(**kwargs)

    def fetch(self, request: HttpRequest):
        log.debug("HTTP request: %s", request)
        try:
            session = self._get_session()
            kwargs = {}
            if request.params is not None:
                kwargs['params'] = request.params
            if request.body is not None:
                kwargs['data'] = request.body
            if request.json is not None:
                kwargs['json'] = request.json
            if request.headers is not None:
                kwargs['headers'] = request.headers
            if request.auth is not None:
                kwargs['auth'] = request.auth
            if request.timeout is not None:
                kwargs['timeout'] = request.timeout
            if request.allow_redirects is not None:
                kwargs['allow_redirects'] = request.allow_redirects
            if request.proxies is not None:
                kwargs['proxies'] = request.proxies
            if request.verify_ssl is not None:
                kwargs['verify'] = request.verify_ssl
            resp = session.request(request.method, request.url, **kwargs)
            resp.raise_for_status()
            response = self._make_response(resp, request)
        except HTTPError as e:
            raise HttpError('{}'.format(e.response),
                            response=self._make_response(e.response, request))
        except Exception as e:
            raise ClientError(e)
        log.debug("HTTP response: %s", response)
        return response

    def _make_response(self, response: Response, request: HttpRequest):
        return HttpResponse(request=request, response=response)

    def _get_session(self):
        try:
            g = gevent.getcurrent().minimal_ident
        except AttributeError:
            g = 0
        if g not in self.sessions:
            self.sessions[g] = self.new_session()
        return self.sessions[g]
